:- module( pack_errors, [pack_errors/0,caught/3,pack_errors_version/2] ).

:- use_module( library(requires) ).
:- lib(options).

/**  <module> Contextual error handling for packs

This is a simple library for (a) middle management handling
of pack errors, (b) provide a simple, uniform way for informing users
where the errors come from, and (c) provide a few useful pre-canned errors.

The main aim is to create contextual version of messages that can be used 
from different packs.

Wrapper errors (each message below is an underlying error)
 * pack_error(Pack,Vrb,Message)
   when Vrb is false there is no message about the pack the error originated from

 * pack_error(Pack,Message)
   as above with Vrb = true

Slowly we adding some generic error handling: see caught/3



 
==
?- throw( pack_error(os,ground_at(3,name(_))) ).
ERROR: pack(os): Ground argument expected at position: 3, but name(_G3369) was found
?- throw( pack_error(os,os_pred/3,ground_at(3,name(_))) ).
ERROR: os:os_pred/3: Ground argument expected at position: 3, but name(_G6625) was found
?- throw( ground_at(3,name(_)) ).
ERROR: Ground argument expected at position: 3, but name(_G1021) was found
==

Prepacked errors: 
	* ground_at(Pos,Pid)
	* lengths_mismatch(Tkn1,Tkn2,Len1,Len2)
	* cast(Term,From,To) )

==
?- throw( pack_error(mlu,lengths_mismatch(learners,predictions,3,4)) ).
ERROR: pack(mlu): Lists for learners and predictions have mismatching lengths: 3 and 4 respectively
?- throw( pack_error(mlu,k_fold_learn/3,lengths_mismatch(learners,predictions,3,4)) ).
ERROR: mlu:k_fold_learn/3: Lists for learners and predictions have mismatching lengths: 3 and 4 respectively
?- throw( pack_error(os,os_term/2,cast(abc('file.csv'),atom)) ).
ERROR: os:os_term/2: Cannot cast: abc(file.csv), to type: atom
==

Defining new pack errors.

example file: 

==
:- multifile( pack_errors:message/3 ).

pack_errors:message( fold_data_insufficient(Dlen,N) ) -->
	['Insufficient length of data (~d) as ~d folds are required'-[Dlen,N]].
pack_errors:message( fold_data_residual(Dlen) ) -->
	['Residual data of length: ~d while splitting folds'-[Dlen]].
==

Once the above has been loaded, try with

==
?- throw( fold_data_insufficient(10,20) ).
ERROR: Insufficient length of data (10) as 20 folds are required
?- throw( pack_error(mlu,fold_data_insufficient(10,20) ) ).
ERROR: pack(mlu): Insufficient length of data (10) as 20 folds are required
?- throw( pack_error(mlu,k_fold_learn/4,fold_data_insufficient(10,20) ) ).
ERROR: mlu:k_fold_learn/4: Insufficient length of data (10) as 20 folds are required
==

@author  nicos angelopoulos
@version 0.1 2016/01/30
@version 0.2 2016/02/24
@tbd     equal length list checking

*/

:- multifile prolog:message//1.

/** caught( Goal, Error, Opts ).

Call Goal with error capture as failure.
On failure, throws packaged Error iff the value of option report(Report) is Report==error.
Option report()  should exist. The only alternative value is Report==fail. 

*/
caught( Goal, _Error, _Opts ) :-
	catch( Goal, _, fail ),
	!.
caught( _Goal, Error, Opts ) :-
	options( report(error), Opts ),
	!,
	throw( Error ).
caught( _Goal, _Error, Opts ) :-
	options( report(fail), Opts ),
	!,
	fail.
caught( _Goal, Error, Opts ) :-
	% fimxe: 
	Mism = opt_mismatch(report,[fail,error],Opts),
	( Error = pack_error(Pac,Pid,_) -> Merr = pack_error(Pac,Pid,Mism); Merr = Mism ),
	throw( Merr ).

/** pack_errors.

This is decoy predicate, writing a simple message.
Here to provide an anchor for documentation pointers.

*/
pack_errors :-
	write( 'Contextual error handling for packs' ), nl.

/** pack_errors_version( -Version, -Date ).

Current version and relase date for the library.

*/
pack_errors_version( 0:2:0, date(2016,2,24) ).

prolog:message(unhandled_exception(pack_error(Pack,Message))) -->
     { debug( pack_errors, 'Unhandled pack_error/2 ~w', Message ) },
	pack_errors:message(from_pack(true,Pack,false)),
     pack_errors:message(Message).
prolog:message(unhandled_exception(pack_error(Pack,Pred,Message))) -->
     { debug( pack_errors, 'Unhandled pack_error/3 ~w', Message ) },
	message(from_pack(true,Pack,Pred)),
     message(Message).
prolog:message(unhandled_exception(pack_error(Pack,Pred,Sil,Message))) -->
     { debug( pack_errors, 'Unhandled pack_error/4 ~w', Message ) },
	message(from_pack(Sil,Pack,Pred)),
     message(Message).
prolog:message(unhandled_exception(Message) ) -->
     { debug( pack_errors, 'Unhandled error, passing as is: ~w', Message ) },
     message(Message).

prolog:message(pack_error(Pack,Message)) -->
     { debug( pack_errors, 'Pack_error/2: ~w', Message ) },
	message(from_pack(true,Pack,false)),
     message(Message).
prolog:message(pack_error(Pack,Pred,Message)) -->
     { debug( pack_errors, 'Pack_error/3: ~w', Message ) },
	message(from_pack(true,Pack,Pred)),
     message(Message).
prolog:message(pack_error(Pack,Pred,Sil,Message)) -->
     { debug( pack_errors, 'Pack_error/4: ~w', Message ) },
	message(from_pack(Sil,Pack,Pred)),
     message(Message).

:- multifile( pack_errors:message/3 ).

% add cuts ? 
%
message( from_pack(false,_Pack,_Pred) ) --> [].
message( from_pack(long,Pack,Pred) ) -->
	pack_errors:message( from_pack_long(Pred,Pack) ).
message( from_pack(true,Pack,Pred) ) -->
	pack_errors:message( from_pack_short(Pred,Pack) ).

message( from_pack_short(false,Pack) ) -->
     { debug( pack_errors, 'From pack short: ~w', Pack ) },
	['pack(~w): '-[Pack] ],
     { debug( pack_errors, 'From pack short: ~w succeeded', Pack ) }.
message( from_pack_short(Pred,Pack) ) -->
	['~w:~w: '-[Pack,Pred] ].
	% ['[~w::~w]: '-[Pack,Pred] ].

message( from_pack_long(false,Pack) ) -->
	['Following error genererated from pack: ~w\n'-[Pack] ].
message( from_pack_long(Pred,Pack) ) -->
	['Following error genererated for predicate: ~w, from pack: ~w'-[Pred,Pack] ].

message( ground_dual_mode(Pos1,Pos2) ) -->  % Pos1 & Pos2 can be lists of positions
	['Ground argument expected either at: ~w, or at: ~w '-[Pos1,Pos2]].
message( ground_at(Pos,Arg) ) -->  % Pid can also be a level up  
	['Ground argument expected at position: ~d, but ~w was found'-[Pos,Arg]].
message( ground_at(Pos,Pid,Arg) ) -->
	['Ground argument expected at position: ~d for predicate: ~w, but, ~w was found'-[Pos,Pid,Arg]].
message( lengths_mismatch(Tkn1,Tkn2,Len1,Len2) ) -->
	['Lists for ~w and ~w have mismatching lengths: ~d and ~d respectively'-[Tkn1,Tkn2,Len1,Len2]].
message( cast(Term,From,To) ) -->
	['Cannot cast: ~w, from type: ~w to type: ~w'-[Term,From,To]].
message( cast(Term,To) ) -->
	['Cannot cast: ~w, to type: ~w'-[Term,To]].

